"""Generic async AWS functions for S3."""

import logging
from typing import Any, Dict, List

from aioradio.aws.utils import AwsServiceManager

LOG = logging.getLogger(__name__)
AWS_SERVICE = AwsServiceManager(service='s3')
S3 = AWS_SERVICE.service_dict


@AWS_SERVICE.active
async def create_bucket(bucket: str) -> Dict[str, str]:
    """Create an s3 bucket.

    Args:
        bucket (str): s3 bucket

    Returns:
        Dict[str, str]: response of operation
    """

    return await S3['client']['obj'].create_bucket(Bucket=bucket)


@AWS_SERVICE.active
async def upload_file(bucket: str, filepath: str, s3_key: str) -> Dict[str, Any]:
    """Upload file to s3.

    Args:
        bucket (str): s3 bucket
        filepath (str): local filepath to upload
        s3_key (str): destination s3 key for uploaded file

    Returns:
        Dict[str, Any]: response of operation
    """

    response = {}
    with open(filepath, 'rb') as fileobj:
        response = await S3['client']['obj'].put_object(Bucket=bucket, Key=s3_key, Body=fileobj.read())

    return response


@AWS_SERVICE.active
async def download_file(bucket: str, filepath: str, s3_key: str):
    """Download file to s3.

    Args:
        bucket (str): s3 bucket
        filepath (str): local filepath for downloaded file
        s3_key (str): s3 key to download
    """

    with open(filepath, 'wb') as fileobj:
        data = await get_object(bucket=bucket, s3_key=s3_key)
        fileobj.write(data)


@AWS_SERVICE.active
async def list_s3_objects(bucket: str, s3_prefix: str, with_attributes: bool=False) -> List[str]:
    """List objects in s3 path.

    Args:
        bucket (str): s3 bucket
        s3_prefix (str): s3 prefix
        with_attributes (bool, optional): return all file attributes in addition to s3 keys. Defaults to False.

    Returns:
        List[str]: [description]
    """

    arr = []
    paginator = S3['client']['obj'].get_paginator('list_objects')
    async for result in paginator.paginate(Bucket=bucket, Prefix=s3_prefix):
        for item in result.get('Contents', []):
            if with_attributes:
                arr.append(item)
            else:
                arr.append(item['Key'])

    return arr


@AWS_SERVICE.active
async def get_s3_file_attributes(bucket: str, s3_key: str) -> Dict[str, Any]:
    """Get s3 objects metadata attributes.

    Args:
        bucket (str): s3 bucket
        s3_key (str): s3 key

    Returns:
        Dict[str, Any]: response of operation
    """

    s3_object = await S3['client']['obj'].get_object(Bucket=bucket, Key=s3_key)
    del s3_object['Body']

    return s3_object


@AWS_SERVICE.active
async def get_object(bucket: str, s3_key: str) -> bytes:
    """Directly download contents of s3 object.

    Args:
        bucket (str): s3 bucket
        s3_key (str): s3 key

    Returns:
        bytes: streaming of s3 key as data bytes
    """

    data = None
    s3_object = await S3['client']['obj'].get_object(Bucket=bucket, Key=s3_key)
    async with s3_object["Body"] as stream:
        data = await stream.read()

    return data


@AWS_SERVICE.active
async def delete_s3_object(bucket: str, s3_prefix: str) -> Dict[str, Any]:
    """Delete object(s) from s3.

    Args:
        bucket (str): s3 bucket
        s3_prefix (str): s3 prefix

    Returns:
        Dict[str, Any]: response of operation
    """

    response = await S3['client']['obj'].delete_object(Bucket=bucket, Key=s3_prefix)

    return response

@AWS_SERVICE.active
async def create_multipart_upload(bucket: str, s3_key: str) -> Dict[str, Any]:
    """Create multipart upload from s3.

    Args:
        bucket (str): s3 bucket
        s3_key (str): s3 prefix

    Returns:
        Dict[str, Any]: response of operation
    """

    response = await S3['client']['obj'].create_multipart_upload(Bucket=bucket, Key=s3_key)

    return response

@AWS_SERVICE.active
async def upload_part(bucket: str, s3_key: str, part: str, part_number: int, upload_id: str) -> Dict[str, Any]:
    """Upload a single part to s3.

    Args:
        bucket (str): s3 bucket
        s3_key (str): s3 prefix
        part (str): the content to be uploaded
        part_number (int): part number
        upload_id (str): multipart upload Id

    Returns:
        Dict[str, Any]: response of operation
    """

    response = await S3['client']['obj'].upload_part(
        Bucket=bucket,
        Key=s3_key,
        Body=part,
        PartNumber=part_number,
        UploadId=upload_id)

    return response

@AWS_SERVICE.active
async def list_parts(bucket: str, s3_key: str, upload_id: str) -> Dict[str, Any]:
    """List parts already uploaded to s3.

    Args:
        bucket (str): s3 bucket
        s3_key (str): s3 prefix
        upload_id (str): multipart upload Id

    Returns:
        Dict[str, Any]: response of operation
    """

    response = await S3['client']['obj'].list_parts(
        Bucket=bucket,
        Key=s3_key,
        UploadId=upload_id)

    return response

@AWS_SERVICE.active
async def complete_multipart_upload(bucket: str, s3_key: str, parts: List, upload_id: str) -> Dict[str, Any]:
    """Complete multipart upload to s3.

    Args:
        bucket (str): s3 bucket
        s3_key (str): s3 prefix
        parts (List): all parts info
        upload_id (str): multipart upload Id

    Returns:
        Dict[str, Any]: response of operation
    """

    response = await S3['client']['obj'].complete_multipart_upload(
        Bucket=bucket,
        Key=s3_key,
        UploadId=upload_id,
        MultipartUpload={'Parts': parts})

    return response

@AWS_SERVICE.active
async def abort_multipart_upload(bucket: str, s3_key: str, upload_id: str) -> Dict[str, Any]:
    """Abort multipart upload.

    Args:
        bucket (str): s3 bucket
        s3_key (str): s3 prefix
        upload_id (str): multipart upload Id

    Returns:
        Dict[str, Any]: response of operation
    """

    response = await S3['client']['obj'].abort_multipart_upload(
        Bucket=bucket,
        Key=s3_key,
        UploadId=upload_id)

    return response




"""Generic async utility functions."""

# pylint: disable=broad-except
# pylint: disable=logging-fstring-interpolation
# pylint: disable=protected-access
# pylint: disable=unnecessary-dunder-call

import asyncio
from asyncio import sleep
from copy import deepcopy
from dataclasses import dataclass, field
from time import time
from types import coroutine
from typing import Any, Dict, List

import aioboto3
import aiobotocore
import aiojobs
import backoff
import botocore


@dataclass
class AwsServiceManager:
    """AWS Service Manager."""

    service: str
    module: str = 'aiobotocore'
    sleep_interval: int = 300
    regions: List[str] = field(default_factory=list)
    scheduler: aiojobs._scheduler = None

    def __post_init__(self):
        services = ['s3', 'sqs', 'secretsmanager', 'dynamodb']
        if self.service not in services:
            raise ValueError(f'service parameter must be one of the following: {services}')

        modules = ['aiobotocore', 'aioboto3']
        if self.module not in modules:
            raise ValueError(f'module parameter must be one of the following: {modules}')

        data = {'client': {'obj': None, 'session': None, 'busy': True}, 'active': 0}
        if self.service == 'dynamodb':
            data['resource'] = {'obj': None, 'session': None, 'busy': True}
        self.service_dict = {region: deepcopy(data) for region in self.regions} if self.regions else data

        try:
            loop = asyncio.get_event_loop()
            if loop and loop.is_running():
                asyncio.create_task(self.create_scheduler())
            else:
                loop.run_until_complete(self.create_scheduler())
        except RuntimeError:
            loop = asyncio.new_event_loop().run_until_complete(self.create_scheduler())

    def add_regions(self, regions: List[str]):
        """Add regions to AWS service.

        Args:
            regions (List[str]): [description]
        """

        data = {'client': {'obj': None, 'session': None, 'busy': True}, 'active': 0}
        if self.service == 'dynamodb':
            data['resource'] = {'obj': None, 'session': None, 'busy': True}

        for region in regions:
            if region not in self.regions:
                self.regions.append(region)
                self.service_dict[region] = deepcopy(data)
                asyncio.get_event_loop().create_task(self.spawn_aws_service_process(region))

    def __del__(self):
        self.scheduler.close()

    async def create_scheduler(self):
        """Schedule jobs."""

        self.scheduler = await aiojobs.create_scheduler()
        if self.regions:
            for region in self.regions:
                await self.spawn_aws_service_process(region)
        else:
            await self.spawn_aws_service_process()

    async def spawn_aws_service_process(self, region: str=''):
        """Spawn scheduler process on a per region basis.

        Args:
            region (str, optional): AWS region. Defaults to ''.
        """

        if region:
            await self.scheduler.spawn(self.aio_server(item='client', region=region))
            if self.module == 'aioboto3':
                await self.scheduler.spawn(self.aio_server(item='resource', region=region))
        else:
            await self.scheduler.spawn(self.aio_server(item='client'))
            if self.module == 'aioboto3':
                await self.scheduler.spawn(self.aio_server(item='resource'))

    async def aio_server(self, item: str, region: str=''):
        """Begin long running server establishing modules service_dict object.

        Args:
            item (str): either 'client' or 'resource' depending on the aws service and python package
            region (str, optional): AWS region. Defaults to ''.
        """

        service_dict = self.service_dict if region == '' else self.service_dict[region]
        await self.establish_client_resource(service_dict[item], item, region=region)

        while True:
            # sleep for defined interval
            await sleep(self.sleep_interval)

            # if all functions are inactive proceed to re-establishing AioSession client
            start = time()
            while service_dict['active'] and (time() - start) < 300:
                await sleep(0.001)

            await self.establish_client_resource(service_dict[item], item=item, region=region, reestablish=True)

    @backoff.on_exception(backoff.expo, botocore.exceptions.ConnectTimeoutError, max_time=120)
    async def establish_client_resource(self, service_dict: Dict[str, Any], item: str, region: str='', reestablish: bool=False):
        """Establish the AioSession client or resource, then re-establish every
        self.sleep_interval seconds.

        Args:
            service_dict (Dict[str, Any]): dict containing info about the service requested
            item (str): either 'client' or 'resource' depending on the aws service and python package
            region (str, optional): AWS region. Defaults to ''.
            reestablish (bool, optional): should async context manager be reinstantiated. Defaults to False.
        """

        kwargs = {'service_name': self.service, 'verify': False}
        if region:
            kwargs['region_name'] = region

        if reestablish:
            service_dict['busy'] = True
            await service_dict['obj'].__aexit__(None, None, None)

        if self.module == 'aiobotocore':
            service_dict['session'] = aiobotocore.session.get_session()
            service_dict['obj'] = await service_dict['session'].create_client(**kwargs).__aenter__()
        elif self.module == 'aioboto3':
            service_dict['session'] = aioboto3.Session()
            func = service_dict['session'].client if item == 'client' else service_dict['session'].resource
            service_dict['obj'] = await func(**kwargs).__aenter__()

        service_dict['busy'] = False

    def get_region(self, args: List[Any], kwargs: Dict[str, Any]) -> str:
        """Attempt to detect the region from the kwargs or args.

        Args:
            args (List[Any]): list of arguments
            kwargs (Dict[str, Any]): Dict of keyword arguments

        Returns:
            str: AWS region
        """

        region = ''

        if 'region' in kwargs:
            region = kwargs['region']
        else:
            for name in args:
                if name in self.regions:
                    region = name
                    break

        return region

    def active(self, func: coroutine) -> Any:
        """Decorator to keep track of currently running functions, allowing the
        AioSession client to only be re-establish when the count is zero to
        avoid functions using a stale client.

        Args:
            func (coroutine): async coroutine

        Returns:
            Any: any
        """

        async def wrapper(*args, **kwargs) -> Any:
            """Decorator wrapper.

            Raises:
                error: some general error during function execution

            Returns:
                Any: any
            """

            obj = self.service_dict[self.get_region(args, kwargs)] if self.regions else self.service_dict

            # Make sure aiobotocore or aioboto3 isn't busy
            while obj['client']['busy'] or ('resource' in obj and obj['resource']['busy']):
                await sleep(0.001)

            result = None
            error = None
            obj['active'] += 1
            try:
                result = await func(*args, **kwargs)
            except Exception as err:
                error = err
            finally:
                obj['active'] -= 1
                if error is not None:
                    raise error

            return result

        return wrapper



"""Long Running Jobs worker script."""

# pylint: disable=broad-except
# pylint: disable=c-extension-no-member
# pylint: disable=logging-fstring-interpolation
# pylint: disable=no-member
# pylint: disable=too-many-instance-attributes

import asyncio
import socket
import traceback
from asyncio import create_task
from dataclasses import dataclass, field
from time import time
from typing import Any, Dict, Tuple
from uuid import uuid4

import httpx
import orjson

from aioradio.aws import sqs
from aioradio.redis import Redis


@dataclass
class LongRunningJobs:
    """Worker that continually pulls from queue (either SQS or Redis list
    implemented like queue), running a job using request parameters conveyed in
    the message.

    Also has functions to send messages to the queue, and to check if
    job is complete.
    """

    redis_host: str = 'localhost'

    # Expiration of cached result. If a job has the same cache_key of a previously run job in redis
    # then we can skip running the job using the previously obtained result.
    expire_cached_result: int = 86400

    # Expiration of cached job data stored as a hashed field.  Example of an ideal TTL is one hour if
    # we expect the job to run in a few seconds/minutes.  If the long running job takes hours
    # then update this to day(s) in seconds.
    expire_job_data: int = 3600

    # Flexibility to define one to many jobs, ex: {"job1_name": (async_func1, 30), "job2_name": (async_func2, 60)}
    # First item of tuple must be an async function that runs your long running job, and the second item the
    # job timeout, which should be a value ~3x or more above the max time a job finishes corresponding to the
    # visibility_timeout if queue_service = 'sqs' or message re-entry into not-started queue if queue_service = 'redis'.
    jobs: Dict[str, Tuple[Any, float]] = field(default_factory=dict)

    # choose between sqs or redis
    queue_service: str = 'sqs'
    # if using sqs than define the queue name and aws region
    sqs_queue: str = None
    sqs_region: str = None

    # If running test cases use fakeredis
    fakeredis: bool = False

    # Trigger the worker to stop running continually
    stop: bool = False

    httpx_client: httpx.AsyncClient = httpx.AsyncClient()

    def __post_init__(self):

        self.queue_service = self.queue_service.lower()
        if self.queue_service not in ['sqs', 'redis']:
            raise ValueError("queue_service must be either 'sqs' or 'redis'.")

        if self.fakeredis:
            self.cache = Redis(fake=True)
        else:
            self.cache = Redis(config={'redis_primary_endpoint': self.redis_host, 'encoding': 'utf-8'})

        self.job_names = set(self.jobs.keys())
        for job_name, job_info in self.jobs.items():
            if len(job_info) != 2:
                raise ValueError('Job info should be provided as a tuple, ex. (job_function, job_timeout)')
            if not isinstance(job_name, str):
                raise ValueError('Job name must be a string value')
            if not isinstance(job_info[1], (int, float)):
                raise ValueError('Job timeout needs to be an integer or float')
            if job_info[1] < 10:
                raise ValueError('Job timeout needs to be at least 10 seconds')
            if job_info[1] > (3600 * 5):
                raise ValueError('Job timeout needs to be no more than 5 hours')

        self.longest_job_timeout = max(i[1] for i in self.jobs.values())
        self.host_uuid = f'{socket.gethostname()}|{uuid4()}'

    async def stop_worker(self):
        """Stop the worker from running continually."""

        self.stop = True

    async def check_job_status(self, uuid: str) -> Dict[str, Any]:
        """Check if the job is done and if so add results to the returned dict.

        Args:
            uuid (str): Unique identifier

        Returns:
            Dict[str Any]: Check job status results
        """

        result = {"uuid": uuid, "job_done": False}
        data = await self.cache.hgetall(key=uuid)
        if data:
            result["job_done"] = data['job_done']
            if result["job_done"]:
                result["results"] = data['results']
        else:
            result["error"] = f"Cannot find {uuid} in Redis Cache"

        return result

    async def send_message(self, job_name: str, params: Dict[str, Any], cache_key: str=None) -> Dict[str, str]:
        """Send message to queue.

        Args:
            job_name (str): Name of job corresponding to a key in self.jobs
            params (Dict[str, Any]): Request parameters needed for job
            cache_key (str, optional): Results cache key. Defaults to None.

        Returns:
            Dict[str, str]: Contains sent message uuid or an error
        """

        if job_name not in self.job_names:
            raise ValueError(f"{job_name} not found in {self.job_names}")

        identifier = str(uuid4())
        items = {"uuid": identifier, "params": params, "cache_key": cache_key, "job_name": job_name}

        result = {}
        try:
            msg = orjson.dumps(items).decode()
            if self.queue_service == 'sqs':
                entries = [{'Id': str(uuid4()), 'MessageBody': msg, 'MessageGroupId': self.host_uuid}]
                await sqs.send_messages(queue=self.sqs_queue, region=self.sqs_region, entries=entries)
            else:
                self.cache.pool.rpush(f'{job_name}-not-started', msg)

            await self.cache.hmset(
                key=identifier,
                items={**items, **{'job_done': False, "results": None}},
                expire=self.expire_job_data
            )
            result['uuid'] = identifier
        except Exception as err:
            result['error'] = str(err)

        return result

    async def start_worker(self):
        """Continually run the worker."""

        while not self.stop:
            try:
                # run job the majority of the time pulling up to 10 messages to process
                for _ in range(10):
                    if self.queue_service == 'sqs':
                        await self.__sqs_pull_messages_and_run_jobs__()
                    else:
                        for job_name in self.job_names:
                            await self.__redis_pull_messages_and_run_jobs__(job_name)

                # verify processing only a fraction of the time
                if self.queue_service == 'redis':
                    for job_name in self.job_names:
                        await self.__verify_processing__(job_name)
            except asyncio.CancelledError:
                print(traceback.format_exc())
                break
            except Exception:
                print(traceback.format_exc())
                await asyncio.sleep(30)

        self.stop = False

    async def __sqs_pull_messages_and_run_jobs__(self):
        """Pull messages one at a time and run job.

        Raises:
            IOError: Redis access failed
        """

        # Since we cannot pull messages for specific jobs we use
        # the longest provided job timeout as the visibility_timeout
        msg = await sqs.get_messages(
            queue=self.sqs_queue,
            region=self.sqs_region,
            wait_time=1,
            visibility_timeout=self.longest_job_timeout,
            max_messages=1,
            attribute_names=['SentTimestamp']
        )

        if not msg:
            await asyncio.sleep(0.1)
        else:
            body = orjson.loads(msg[0]['Body'])
            key = body['cache_key']
            job_name = body['job_name']

            callback_url = body['params'].get('callback_url', '')
            body['params'].pop('callback_url', None)

            data = None if key is None else await self.cache.get(key)
            if data is None:
                # No results found in cache so run the job
                data = await self.jobs[job_name][0](body['params'])

                # Set the cached parameter based key with results
                if key is not None and not await self.cache.set(key, data, expire=self.expire_cached_result):
                    raise IOError(f"Setting cache string failed for cache_key: {key}")

            # Send results via POST request if necessary
            if callback_url:
                create_task(self.httpx_client.post(callback_url, json={'results': data, 'uuid': body['uuid']}, timeout=30))

            # Update the hashed UUID with processing results
            await self.cache.hmset(
                key=body['uuid'],
                items={**body, **{'results': data, 'job_done': True}},
                expire=self.expire_job_data
            )

            entries = [{'Id': str(uuid4()), 'ReceiptHandle': msg[0]['ReceiptHandle']}]
            await sqs.delete_messages(queue=self.sqs_queue, region=self.sqs_region, entries=entries)

            total = round(time() - float(msg[0]['Attributes']['SentTimestamp'])/1000, 3)
            print(f"Async college match processing time for UUID {body['uuid']} took {total} seconds")

    async def __redis_pull_messages_and_run_jobs__(self, job_name):
        """Pull messages one at a time and run job.

        Args:
            job_name (str): Name of job corresponding to a key in self.jobs

        Raises:
            IOError: Redis access failed
        """

        # in the future convert lpop to lmove and also look into integrating with async aioredis
        msg = self.cache.pool.lpop(f'{job_name}-not-started')
        if not msg:
            await asyncio.sleep(0.1)
        else:
            body = orjson.loads(msg)
            key = body['cache_key']

            callback_url = body['params'].get('callback_url', '')
            body['params'].pop('callback_url', None)

            # Add start time and push msg to <self.name>-in-process
            body['start_time'] = time()
            self.cache.pool.rpush(f'{job_name}-in-process', orjson.dumps(body).decode())

            data = None if key is None else await self.cache.get(key)
            if data is None:
                # No results found in cache so run the job
                data = await self.jobs[job_name][0](body['params'])

                # Set the cached parameter based key with results
                if key is not None and not await self.cache.set(key, data, expire=self.expire_cached_result):
                    raise IOError(f"Setting cache string failed for cache_key: {key}")

            # Send results via POST request if necessary
            if callback_url:
                create_task(self.httpx_client.post(callback_url, json={'results': data, 'uuid': body['uuid']}, timeout=30))

            # Update the hashed UUID with processing results
            await self.cache.hmset(
                key=body['uuid'],
                items={**body, **{'results': data, 'job_done': True}},
                expire=self.expire_job_data
            )

    async def __verify_processing__(self, job_name):
        """Verify processing completed fixing issues related to app crashing or
        scaling down servers when using queue_service = 'redis'.

        Args:
            job_name (str): Name of job corresponding to a key in self.jobs
        """

        timeout = self.jobs[job_name][1]
        for _ in range(10):
            msg = self.cache.pool.lpop(f'{job_name}-in-process')
            if not msg:
                break

            body = orjson.loads(msg)
            job_done = await self.cache.hget(key=body['uuid'], field='job_done')
            if job_done is None:
                pass    # if the cache is expired then we can typically ignore doing anything
            elif not job_done:
                if (time() - body['start_time']) > timeout:
                    print(f'{job_name} with uuid: {body["uuid"]} failed after {timeout} seconds. \
                            Pushing msg back to {job_name}-not-started.')
                    self.cache.pool.rpush(f'{job_name}-not-started', msg)
                else:
                    self.cache.pool.rpush(f'{job_name}-in-process', msg)

    @staticmethod
    async def build_cache_key(params: Dict[str, Any], separator='|') -> str:
        """Build a cache key from a dictionary object.  Concatenate and
        normalize key-values from an unnested dict, taking care of sorting the
        keys and each of their values (if a list).

        Args:
            params (Dict[str, Any]): dict object to use to build cache key
            separator (str, optional): character to use as a separator in the cache key. Defaults to '|'.
        Returns:
            str: dict object converted to string
        """

        keys = sorted(params.keys())
        concat_key = separator.join([
            f'{k}={orjson.dumps(params[k], option=orjson.OPT_SORT_KEYS).decode()}'.replace('"', '')
            for k in keys if params[k] != [] and params[k] is not None
        ])

        return concat_key
